@@ -59,11 +59,16 @@ class Simulation:
 
     def normalize(self) -> None:
         new_frames = []
+        first = True
         for frame in self.frames:
             aspect = frame.w / frame.h
+
             new_domain = DomainParams(
                 frame.n, sqrt(frame.n * aspect), sqrt(frame.n / aspect), frame.r
             )
+            if first:
+                self.domain = new_domain
+                first = False
 
             new_points = frame.site_arr * np.array(
                 [new_domain.w / frame.w, new_domain.h / frame.h]
@@ -159,6 +164,9 @@ class Simulation:
         sim, frames = Simulation.load(path)
         for frame in frames:
             sim.frames.append(sim.energy.mode(*frame["domain"], frame["arr"]))
+            for k, v in frame["stats"].items():
+                if k not in sim.frames[-1].stats:
+                    sim.frames[-1].stats[k] = v
 
         return sim
 
@@ -233,13 +241,14 @@ class Flow(Simulation):
             change, grad = frame.iterate(self.step_size)
 
             new_frame = self.energy.mode(*self.domain, frame.add_sites(change))
-            grad_norm = np.linalg.norm(grad)
             end = timer()
 
+            grad_norm = np.linalg.norm(grad) / (self.domain.n ** 0.5)
+
             if self.adaptive:
                 error = change - grad * self.step_size
                 tol = 10 ** min(-3, -2 + log10(grad_norm))
-
+                # tol = 10 ** -10
                 self.step_size *= (tol / np.linalg.norm(error)) ** 0.5
 
             if not save:
@@ -315,7 +324,7 @@ class Search(Simulation):
         new_sites: Optional[numpy.ndarray] = None,
     ) -> None:
         if log:
-            print(f"Travel - {self.domain}", flush=True)
+            print(f"Search - {self.domain} - {len(self)}", flush=True)
         if save and len(self) == 0:
             self.save(self.initial_data, True)
 
@@ -328,16 +337,15 @@ class Search(Simulation):
             sim.run(False, log, log_steps)
 
             self.frames.append(sim[-1])
+            # Get Hessian,and check nullity. If > 2, perturb.
+            hess = self.frames[i].hessian
+            eigs = np.sort(np.linalg.eig(hess)[0])
+            self.frames[i].stats["eigs"] = eigs
+
             if save:
                 self.save(self.frame_data(i))
             if log:
                 print(f"Equilibrium: {i:04}\n", flush=True)
-
-            # Get Hessian,and check nullity. If > 2, perturb.
-            hess = self.frames[i].hessian(10e-5)
-            eigs = np.sort(np.linalg.eig(hess)[0])
-            self.frames[i].stats["eigs"] = eigs
-
             zero_eigs = np.count_nonzero(
                 np.isclose(eigs, np.zeros((len(eigs),)), atol=1e-4)
             )
@@ -386,10 +394,7 @@ class Shrink(Simulation):
     ) -> None:
         super().__init__(domain, energy, name=name)
         self.step_size, self.thres, self.adaptive = step_size, thres, adaptive
-        self.delta, self.stop_width = (
-            self.domain.w * delta / 100,
-            self.domain.w * stop_width,
-        )
+        self.delta, self.stop_width = delta, stop_width
 
     @property
     def initial_data(self) -> Dict:
@@ -418,8 +423,19 @@ class Shrink(Simulation):
             self.save(self.initial_data, True)
 
         width = self.domain.w if len(self.frames) == 0 else self.frames[-1].w
+        if len(self.frames) != 0:
+            new_sites = self.frames[-1].site_arr
+            width = self.frames[-1].w - self.delta
+        else:
+            width = self.domain.w
         i = 0
-        while width >= self.stop_width:
+
+        if delta < 0:
+            cond = width <= self.stop_width
+        if delta > 0:
+            cond = width <= self.stop_width
+
+        while cond:
             # Get to equilibrium.
             new_domain = DomainParams(
                 self.domain.n, width, self.domain.h, self.domain.r
